"""Call the main functions. Used by `poetry` as main script."""

from psa_prizes.analyze import analyze
from psa_prizes.scrape import scrape


def main() -> None:
    """Run psa-prizes."""
    scrape()
    analyze()



"""Scrape data from PSA and save it to one csv-file for each card."""

import math
import os
import time
from typing import Any, List, Tuple, Union

from bs4 import BeautifulSoup
import pandas as pd
import requests


def scrape() -> None:
    """Parse urls to `scrape_psa_prizes` and store as .csv-files."""
    # Read in urls from urls.txt.
    if not os.path.exists("input/input.csv"):
        raise ValueError("'input.csv' not found")

    df = pd.read_csv("input/input.csv", sep=";")
    urls = df["link (str)"].tolist()

    # If psa-prizes/output/data doesn't exist, create it.
    if not os.path.exists("output/data"):
        os.makedirs("output/data")

    # Iterate over all urls.
    for url in urls:
        # Execute web scraping.
        scrape_psa_prizes(url)


def scrape_psa_prizes(card_url: str) -> None:
    """Scrape data as specified by "link (str)" in `input.csv`."""
    print("collecting data for {}".format(card_url))

    # Get html data from input url.
    sess = requests.Session()
    sess.mount("https://", requests.adapters.HTTPAdapter(max_retries=5))
    r = sess.get(card_url)
    r.raise_for_status()
    soup = BeautifulSoup(r.text, "html5lib")
    time.sleep(5)

    # Get image url links.
    images = _get_image_urls(soup)

    # Get sale prices.
    prizes = _get_prizes(soup)

    # Get dates of the sales.
    dates = _get_sale_dates(soup)

    # Get PSA grades and qualifiers.
    grades, quals = _get_grades(soup)

    # Get lot url links.
    lots = _get_lot_urls(soup)

    # Get auction houses.
    a_houses = _get_auction_houses(soup)

    # Get names of the sellers.
    sellers = _get_seller_names(soup)

    # Get sale types (auction, BIN, Best Offer, etc).
    sale_types = _get_sale_types(soup)

    # Get PSA certification numbers.
    certs = _get_psa_certs(soup)

    # Create a dataframe.
    df = pd.DataFrame(
        {
            "date": dates,
            "grade": grades,
            "qualifier": quals,
            "prize": prizes,
            "auction_house": a_houses,
            "seller": sellers,
            "sale_type": sale_types,
            "psa_certification": certs,
            "img_url": images,
            "lot_url": lots,
        }
    )

    # Write to Excel file.
    df.to_csv(_get_file_name(card_url), index=False)


def _get_image_urls(soup: Any) -> List[Union[str, float]]:  # noqa: D102
    image_data = [n for n in soup.find_all("div", {"class": "item-image"})]
    images: List[Union[str, float]] = []
    for n in image_data:
        html = str(n)
        if "href" not in html:
            images.append(math.nan)
            continue
        images.append(html.split('href="')[1].split('"')[0])
    return images


def _get_prizes(soup: Any) -> List[float]:  # noqa: D102
    prices = [
        float(n.string.strip("$").replace(",", ""))
        for n in soup.find_all("div", {"class": "item item-price"})
    ]

    return prices


def _get_sale_dates(soup: Any) -> List[str]:  # noqa: D102
    dates = [n.string for n in soup.find_all("div", {"class": "item item-date"})]
    return dates


def _get_grades(
    soup: Any
) -> Tuple[List[Union[str, float]], List[Union[str, float]]]:  # noqa: D102
    """Prefer grades as floats not strings. One str converts whole col to str."""
    grade_data = soup.find_all("div", {"class": "item item-grade"})
    grades: List[Union[str, float]] = []
    quals: List[Union[str, float]] = []
    for n in grade_data:
        html = str(n)
        try:
            grades.append(float(html.split("</span>")[1].split("<")[0].strip()))
        except ValueError:
            grades.append(html.split("</span>")[1].split("<")[0].strip())
        if "<strong>" in html:
            quals.append(html.split("<strong>")[1].split("<")[0].strip())
        else:
            quals.append(math.nan)

    return grades, quals


def _get_lot_urls(soup: Any) -> List[Union[str, float]]:  # noqa: D102
    lot_data = soup.find_all("div", {"class": "item item-lot"})
    base_url = "https://www.psacard.com"
    lots: List[Union[str, float]] = []
    for n in lot_data:
        html = str(n)
        if "href" not in html:
            lots.append(math.nan)
            continue
        lots.append(base_url + html.split('href="')[1].split('"')[0])
    return lots


def _get_auction_houses(soup: Any) -> List[str]:  # noqa: D102
    a_houses = [
        n.string for n in soup.find_all("div", {"class": "item item-auctionhouse"})
    ]
    return a_houses


def _get_seller_names(soup: Any) -> List[str]:  # noqa: D102
    sellers = [
        n.string for n in soup.find_all("div", {"class": "item item-auctionname"})
    ]
    return sellers


def _get_sale_types(soup: Any) -> List[str]:  # noqa: D102
    sale_types = [
        n.string for n in soup.find_all("div", {"class": "item item-auctiontype"})
    ]
    return sale_types


def _get_psa_certs(soup: Any) -> List[int]:  # noqa: D102
    certs = [
        int(str(n).split("</span>")[1].split("<")[0])
        for n in soup.find_all("div", {"class": "item item-cert"})
    ]
    return certs


def _get_file_name(card_url: str) -> str:  # noqa: D102
    f_name = card_url.split("-cards/")[1].split("/values")[0].replace("/", "-")
    return "{}.csv".format(os.path.join("output/data", f_name))



"""Analysis tools."""
import ast
import json
import os
from typing import Any, Dict, Optional, Union

import matplotlib.dates as dates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd


def analyze() -> None:
    """Return info messages as dict, plot prize timeline and save both.

    Notes
    -----
    For some cards, the grade data is of type `str` instead of `int` or `float`.
    pd.DataFrames assign the data-types column-wise by using the most compatible
    type. For example, the grade "Authentic" for only one card causes pandas
    to transform the whole column with grades in  {1.0, 1.5, ... 10.0} to be
    stored as `str`. Therefore, all grades are converted to `str` for a unified
    treatment. These strings are written like floats with one decimal place.

    """
    # Read grades. Delimeter has to be semicolon only.
    if not os.path.exists("input/input.csv"):
        raise ValueError("'input.csv' not found")
    df = pd.read_csv("input/input.csv", sep=";")

    # Convert List-String in input column to List[Union[float,int]].
    grades = df["grades (list)"].apply(ast.literal_eval).tolist()

    # Iterate over cards in input DataFrame.
    for row in range(len(df)):
        # Get card_name from urls
        card_name = _get_card_name(df.iloc[row, 0])

        # Important: Convert col to `str` as described in the notes.
        df_out = pd.read_csv(f"output/data/{card_name}.csv", parse_dates=["date"])
        df_out["grade"] = df_out["grade"].astype(str)

        # Iterate over grades per card.
        for g in grades[row]:

            # Parse grade as str due to reasons above.
            g = str(float(g))

            # Info on grades outside {1.0, 1.5, ... 10.0}.
            msg_grades = _str_unusual_grades(df_out)

            # Info on the share with this grade.
            msg_grades_n = _str_n_grade(df_out, g)

            # Drop rows with different grades.
            df_filt = _filter_grade_data(df_out, g)

            # Compute compund annual growth rate (CAGR).
            msg_return = _str_comp_ann_g(df_filt)

            # Store infos in dictionary and print.
            card_dict: Dict[str, Optional[str]] = {
                "ident": f"{card_name}-{g}",
                "compound annual growth": msg_return,
                "info grades number": msg_grades_n,
            }
            # Drop message if there are no unausual grades.
            if msg_grades_n is None:
                pass
            else:
                card_dict["info grades"] = msg_grades
            print(card_dict)

            # Save dictionary.
            if not os.path.exists("output/nmbrs"):
                os.makedirs("output/nmbrs")
            with open(f"output/nmbrs/{card_name}-grade-{g}.json", "w") as fp:
                json.dump(card_dict, fp)

            # Plot and save prize trend.
            _scatter_prize_time(df_filt, f"{card_name}-grade-{g}")


def _str_unusual_grades(df: pd.DataFrame) -> Union[str, None]:
    """Print the number of unusual grades."""
    grades = np.arange(0, 10.5, 0.5).astype(float)
    catch_grades = []
    for item in df["grade"]:
        try:
            if float(item) not in grades:
                catch_grades.append(item)
        except ValueError:
            catch_grades.append(item)

    if catch_grades == []:
        return None
    else:
        return (
            f"– Over all grades, {len(catch_grades)} of {len(df)} cards do not receive"
            f" standard grades. These grades are in {set(catch_grades)}"
        )


def _get_card_name(card_url: str) -> str:  # noqa: D102
    c_name = card_url.split("-cards/")[1].split("/values")[0].replace("/", "-")
    return c_name


def _str_n_grade(df: pd.DataFrame, grade: str) -> str:
    """Print the number of cards with grade `grade`."""
    n_cards = len(df)
    n_grade = len(df[(df["grade"]) == grade])
    perc = round((n_grade / n_cards) * 100, 2)

    return (
        f"– The number of cards with grade {grade} is {n_grade} "
        f"of {n_cards} cards. That is {perc}%."
    )


def _filter_grade_data(df: pd.DataFrame, grade: str) -> pd.DataFrame:
    """Reduce df to date and price data for cards with grade `grade`."""
    df = df[(df["grade"]) == grade]
    df = df[["date", "prize"]]

    return df


def _str_comp_ann_g(df: pd.DataFrame) -> str:
    """Print the average annual prize growth."""
    if df.empty is True:
        return "There is no prize data for this grade."
    else:
        df["year"] = pd.DatetimeIndex(df["date"]).year
        df["avg_prize_in_year"] = df.groupby("year")["prize"].transform("mean")

        # Create pd.DataFrame for annual average prizes.
        years = df["year"].drop_duplicates()
        prizes = df["avg_prize_in_year"].drop_duplicates()
        avg_df = pd.DataFrame({"year": years, "prize": prizes})

        # cagr = (1 + R) ** (1 / n) - 1
        t_0 = min(avg_df["year"])
        t_T = max(avg_df["year"])
        p_0 = avg_df[t_0 == avg_df["year"]]["prize"].iloc[0]
        p_T = avg_df[t_T == avg_df["year"]]["prize"].iloc[0]
        R = p_T / p_0
        n = t_T - t_0
        cagr = R ** (1 / n) - 1

        # Compute compound annual growth rate in percent.
        cagr = round(cagr * 100, 2)

        return (
            f"– The compound annual growth rate from {min(years)} "
            f"to {max(years)} is {cagr}%."
        )


def _scatter_prize_time(df: pd.DataFrame, title: str) -> Any:
    if df.empty is True:
        print("No prize data, no plot.")
    else:
        x = dates.date2num(df["date"])
        y = df["prize"]

        fig, ax = plt.subplots(figsize=(12, 9))

        ax.scatter(df["date"], y, alpha=0.66)

        # Draw red trend line.
        fit = np.polyfit(x, y, deg=3)
        p = np.poly1d(fit)
        ax.plot(x, p(x), "r--")

        # Rotate date labels.
        fig.autofmt_xdate()

        ax.grid(linestyle="-", color="black", alpha=0.25)
        ax.tick_params(length=6, width=2, labelsize=20)
        ax.set_title(title, fontsize=22)
        ax.set_xlabel(xlabel="Date", size=26, labelpad=14)
        ax.set_ylabel(ylabel="Prize in $", size=26, labelpad=14)

        fig.tight_layout()

        if not os.path.exists("output/img"):
            os.makedirs("output/img")

        plt.savefig(f"output/img/{title}.png")
        plt.show()

        return fig
